/* beginnings of a PDF parser in hammer
 * pesco 2019
 */

#include <string.h>	/* strncmp(), memset() */

#include <hammer/hammer.h>
#include <hammer/glue.h>

/* convenience macros */
#define SEQ(...)	h_sequence(__VA_ARGS__, NULL)
#define CHX(...)	h_choice(__VA_ARGS__, NULL)
#define REP(P,N)	h_repeat_n(P, N)
#define IGN(P)		h_ignore(P)
#define LIT(S)		h_ignore(h_literal(S))
#define IN(STR)		h_in(STR, sizeof(STR) - 1)
#define NOT_IN(STR)	h_not_in(STR, sizeof STR - 1)


/*
 * some helpers
 */

HParser *p_fail;
HParser *p_epsilon;
HParser *p_return_0;
HParser *p_return_1;

/* a combinator to parse a given character but return a different value */

HParsedToken *
act_return_uint(const HParseResult *p, void *u)
{
	return H_MAKE_UINT((uint64_t)(uintptr_t)u);
}

HParser *
p_mapch(uint8_t c, uint8_t v)
{
	return h_action(h_ch(c), act_return_uint, (void *)(uintptr_t)v);
}

/* a parser that just returns a given token */

HParsedToken *
act_return(const HParseResult *p, void *u)
{
	return u;
}

HParser *
p_return__m(HAllocator *mm__, const HParsedToken *tok)
{
	return h_action__m(mm__, p_epsilon, act_return, (void *)tok);
}

/* a parser that just returns a given uint */
HParser *
p_return_uint__m(HAllocator *mm__, uint64_t x)
{
	return h_action__m(mm__, p_epsilon, act_return_uint, (void *)x);
}

/* a helper to compare an HBytes to a string */
bool
bytes_eq(HBytes b, const char *s)
{
	return strncmp(s, b.token, b.len) == 0 && b.len == strlen(s);
}

/* a helper to look up a value in a dictionary */
const HParsedToken *
dictentry(const HCountedArray *dict, const char *key)
{
	HParsedToken *ent;
	HBytes k;
	size_t len;

	if (dict == NULL)
		return NULL;

	len = strlen(key);
	for (size_t i = 0; i < dict->used; i++) {
		ent = dict->elements[i];
		k = H_INDEX_BYTES(ent, 0);

		if (k.len == len && bytes_eq(k, key))
			return H_INDEX_TOKEN(ent, 1);
	}

	return NULL;
}

bool
validate_eq_uint(HParseResult *p, void *u)
{
	const HParsedToken *v = p->ast;
	return (v != NULL && v->token_type == TT_UINT &&
	    v->uint == (uint64_t)(uintptr_t)u);
}


/*
 * auxiliary global data structure needed by the parser
 */
struct Env {
	const char *infile;
	const char *input;
	size_t sz;

	const HParsedToken **xrefs;	/* all xref sections of the file */
	size_t nxrefs;
};


/*
 * custom token types
 */
HTokenType TT_XREntry, TT_Ref;

typedef struct {
	enum {XR_FREE, XR_INUSE, XR_OBJSTM} type;
	union {
		struct { size_t next, ngen; } f;	/* free */
		struct { size_t offs, gen; } n;		/* inuse */
		struct { size_t stm, idx; } o;		/* objstm */
	};
	const HParsedToken *obj;
} XREntry;

typedef struct { size_t nr, gen; } Ref;

void
pp_xrentry(FILE *stream, const HParsedToken *tok, int indent, int delta)
{
	XREntry *xr = H_CAST(XREntry, tok);

	switch (xr->type) {
	case XR_FREE:
		fprintf(stream, "{ \"type\":\"free\", \"next\":%zu, "
		    "\"ngen\":%zu }", xr->f.next, xr->f.ngen);
		break;
	case XR_INUSE:
		fprintf(stream, "{ \"type\":\"inuse\", \"offs\":%zu, "
		    "\"gen\":%zu }", xr->n.offs, xr->n.gen);
		break;
	case XR_OBJSTM:
		fprintf(stream, "{ \"type\":\"objstm\", \"stm\":%zu, "
		    "\"idx\":%zu }", xr->o.stm, xr->o.idx);
		break;
	default:
		assert(!"reached");
	}
}

void
pp_ref(FILE *stream, const HParsedToken *tok, int indent, int delta)
{
	Ref *r = H_CAST(Ref, tok);

	fprintf(stream, "[%zu,%zu]", r->nr, r->gen);
}


/*
 * semantic actions
 */

HParsedToken *
act_digit(const HParseResult *p, void *u)
{
	return H_MAKE_UINT(H_CAST_UINT(p->ast) - '0');
}
#define act_pdigit act_digit
#define act_odigit act_digit

HParsedToken *
act_hlower(const HParseResult *p, void *u)
{
	return H_MAKE_UINT(H_CAST_UINT(p->ast) - 'a');
}

HParsedToken *
act_hupper(const HParseResult *p, void *u)
{
	return H_MAKE_UINT(H_CAST_UINT(p->ast) - 'A');
}

HParsedToken *
act_nat(const HParseResult *p, void *u)
{
	uint64_t x = 0;
	HCountedArray *seq = H_CAST_SEQ(p->ast);

	// XXX check for overflow
	for (size_t i = 0; i < seq->used; i++)
		x = x*10 + H_CAST_UINT(seq->elements[i]);

	return H_MAKE_UINT(x);
}
#define act_xrnat act_nat
#define act_xroff act_nat
#define act_xrgen act_nat

HParsedToken *
act_pnat(const HParseResult *p, void *u)
{
	uint64_t x = H_FIELD_UINT(0);
	HCountedArray *seq = H_FIELD_SEQ(1);

	// XXX check for overflow
	for (size_t i = 0; i < seq->used; i++)
		x = x*10 + H_CAST_UINT(seq->elements[i]);
	
	return H_MAKE_UINT(x);
}

HParsedToken *
act_intg(const HParseResult *p, void *u)
{
	int64_t x = 0;
	HCountedArray *seq = H_FIELD_SEQ(1);

	// XXX check for overflow
	for (size_t i = 0; i < seq->used; i++)
		x = x*10 + H_CAST_UINT(seq->elements[i]);

	HParsedToken *sgn = H_INDEX_TOKEN(p->ast, 0);
	if (sgn->token_type == TT_BYTES &&
	    sgn->bytes.token[0] == '-')
		x = -x;
	
	return H_MAKE_SINT(x);
}

HParsedToken *
act_real(const HParseResult *p, void *u)
{
	double x = 0;
	double f = 0;
	HCountedArray *whole = H_FIELD_SEQ(1, 0);
	HCountedArray *fract = H_FIELD_SEQ(1, 2);

	// XXX check for overflow
	for (size_t i = 0; i < whole->used; i++)
		x = x*10 + H_CAST_UINT(whole->elements[i]);
	for (size_t i = 0; i < fract->used; i++)
		f = (f + H_CAST_UINT(fract->elements[fract->used - 1 - i])) / 10;

	HParsedToken *sgn = H_INDEX_TOKEN(p->ast, 0);
	if (sgn->token_type == TT_BYTES &&
	    sgn->bytes.token[0] == '-')
		x = -x;
	
	return H_MAKE_UINT(x + f);	// XXX H_MAKE_DOUBLE (-> pprint)
}

HParsedToken *
act_ref(const HParseResult *p, void *u)
{
	Ref *r = H_ALLOC(Ref);

	r->nr = H_FIELD_UINT(0);
	r->gen = H_FIELD_UINT(1);

	return H_MAKE(Ref, r);
}

HParsedToken *
act_token(const HParseResult *p, void *u)
{
	HCountedArray *seq = H_CAST_SEQ(p->ast);
	uint8_t *bytes;

	bytes = h_arena_malloc(p->arena, seq->used);
	for (size_t i = 0; i < seq->used; i++)
		bytes[i] = H_CAST_UINT(seq->elements[i]);

	return H_MAKE_BYTES(bytes, seq->used);
}
#define act_nstr act_token

HParsedToken *
act_nesc(const HParseResult *p, void *u)
{
	return H_MAKE_UINT(H_FIELD_UINT(1)*16 + H_FIELD_UINT(2));
}

#define act_schars h_act_flatten
#define act_string act_token

HParsedToken *
act_octal(const HParseResult *p, void *u)
{
	uint64_t x = 0;
	HCountedArray *seq = H_CAST_SEQ(p->ast);

	for (size_t i = 0; i < seq->used; i++)
		x = x*8 + H_CAST_UINT(seq->elements[i]);

	return H_MAKE_UINT(x);
}

HParsedToken *
act_xrent(const HParseResult *p, void *u)
{
	XREntry *xr = H_ALLOC(XREntry);
	char c = H_FIELD_UINT(2);

	xr->obj = NULL;
	switch (c) {
	case 'f':
		xr->type = XR_FREE;
		xr->f.next = H_FIELD_UINT(0);
		xr->f.ngen = H_FIELD_UINT(1);
		break;
	case 'n':
		xr->type = XR_INUSE;
		xr->n.offs = H_FIELD_UINT(0);
		xr->n.gen = H_FIELD_UINT(1);
		break;
	default:
		assert(!"reached");
	}

	return H_MAKE(XREntry, xr);
}

HParsedToken *
act_xrstment(const HParseResult *p, void *u)
{
	XREntry *xr = H_ALLOC(XREntry);

	xr->obj = NULL;
	xr->type = H_FIELD_UINT(0);
	switch (xr->type) {
	case XR_FREE:
		xr->f.next = H_FIELD_UINT(1);
		xr->f.ngen = H_FIELD_UINT(2);
		break;
	case XR_INUSE:
		xr->n.offs = H_FIELD_UINT(1);
		xr->n.gen = H_FIELD_UINT(2);
		break;
	case XR_OBJSTM:
		xr->o.stm = H_FIELD_UINT(1);
		xr->o.idx = H_FIELD_UINT(2);
		break;
	default:
		assert(!"reached");
	}

	return H_MAKE(XREntry, xr);
}

#define act_xrefs h_act_last

HParsedToken *
act_rest(const HParseResult *p, void *env)
{
	struct Env *aux = env;
	size_t offset = H_CAST_UINT(p->ast) / 8;

	return H_MAKE_BYTES(aux->input + offset, aux->sz - offset);
}

/* stream semantics (defined further below) */
bool validate_xrstm(HParseResult *, void *);
HParsedToken *act_xrstm(const HParseResult *, void *);


/*
 * input grammar
 */

HParser *p_pdf;
HParser *p_pdfdbg;
HParser *p_startxref;
HParser *p_xref;
HParser *p_objdef;

/* continuations for h_bind() */
HParser *kstream(HAllocator *, const HParsedToken *, void *);
HParser *kxstream(HAllocator *, const HParsedToken *, void *);

void
init_parser(struct Env *aux)
{
	TT_XREntry =	h_allocate_token_new("XREntry", NULL, pp_xrentry);
	TT_Ref =	h_allocate_token_new("Ref", NULL, pp_ref);

	/* lines */
	H_RULE(cr,	p_mapch('\r', '\n'));	/* semantic value: \n */
	H_RULE(lf,	h_ch('\n'));		/* semantic value: \n */
	H_RULE(crlf,	h_right(cr, lf));	/* semantic value: \n */
	H_RULE(eol,	CHX(crlf, cr, lf));
	H_RULE(line,	h_many(NOT_IN("\r\n")));

	/* character classes */
#define LWCHARS	"\0\t\f "
#define WCHARS	LWCHARS "\n\r"
#define DCHARS	"()<>[]{}/%"
	H_RULE(wchar,	IN(WCHARS));			/* white-space */
	H_RULE(lwchar,	IN(LWCHARS));			/* "line" whitespace */
	//H_RULE(dchar,	IN(DCHARS));			/* delimiter */
	H_RULE(rchar,	NOT_IN(WCHARS DCHARS));		/* regular */
	H_RULE(nchar,	NOT_IN(WCHARS DCHARS "#"));	/* name */
	H_RULE(schar,	NOT_IN("()\n\r\\"));		/* string literal */
	H_ARULE(digit,	h_ch_range('0', '9'));
	H_ARULE(pdigit,	h_ch_range('1', '9'));
	H_ARULE(hlower,	h_ch_range('a', 'f'));
	H_ARULE(hupper,	h_ch_range('A', 'F'));
	H_RULE(hdigit,	CHX(digit, hlower, hupper));
	H_ARULE(odigit,	h_ch_range('0', '7'));
	H_RULE(sign,	IN("+-"));

	H_RULE(sp,	h_ch(' '));
	H_RULE(percent,	h_ch('%'));
	H_RULE(period,	h_ch('.'));
	H_RULE(slash,	h_ch('/'));
	H_RULE(hash,	h_ch('#'));
	H_RULE(bslash,	h_ch('\\'));
	H_RULE(lparen,	h_ch('('));
	H_RULE(rparen,	h_ch(')'));
	H_RULE(langle,	h_ch('<'));
	H_RULE(rangle,	h_ch('>'));
	H_RULE(lbrack,	h_ch('['));
	H_RULE(rbrack,	h_ch(']'));

	/* whitespace */
	H_RULE(comment,	SEQ(percent, line));
	H_RULE(ws,	IGN(h_many(CHX(wchar, comment))));
	H_RULE(lws,	IGN(h_many(lwchar)));

#define TOK(X)	h_middle(ws, X, h_not(rchar))
#define KW(S)	TOK(LIT(S))
#define TOKD(X)	h_right(ws, X)	/* for tokens that end on delimiters */
// XXX this allows, for instance, "<<<<" to be parsed as "<< <<". ok?

	/* misc */
	H_RULE(nl,	IGN(h_right(lws, eol)));
	H_RULE(end,	h_end_p());
	H_RULE(epsilon,	h_epsilon_p());
	H_RULE(empty,	SEQ(epsilon));
	H_ARULE(nat,	TOK(h_many1(digit)));
	H_ARULE(pnat,	TOK(SEQ(pdigit, h_many(digit))));

#define OPT(X)	CHX(X, epsilon)

	/*
	 * objects
	 */
	
	H_ARULE(ref,	SEQ(pnat, nat, KW("R")));
	H_RULE(null,	KW("null"));
	H_RULE(boole,	CHX(KW("true"), KW("false")));

	/* numbers */
	H_RULE(digits,	h_many1(digit));
	H_ARULE(intg,	TOK(SEQ(h_optional(sign), digits)));
	H_RULE(realnn,	CHX(SEQ(digits, period, digits),	/* 12.3 */
			    SEQ(digits, period, empty),		/* 123. */
			    SEQ(empty, period, digits)));	/* .123 */
	H_ARULE(real,	TOK(SEQ(h_optional(sign), realnn)));

	/* names */
	H_ARULE(nesc,	SEQ(hash, hdigit, hdigit));
	H_ARULE(nstr,	h_many(CHX(nchar, nesc)));	/* '/' is valid */
	H_RULE(name,	TOK(h_right(slash, nstr)));

	/* strings */
	H_RULE(snest,	h_indirect());
	H_RULE(bsn,	p_mapch('n', 0x0a));	/* LF */
	H_RULE(bsr,	p_mapch('r', 0x0d));	/* CR */
	H_RULE(bst,	p_mapch('t', 0x09));	/* HT */
	H_RULE(bsb,	p_mapch('b', 0x08));	/* BS (backspace) */
	H_RULE(bsf,	p_mapch('f', 0x0c));	/* FF */
	H_RULE(escape,	CHX(bsn, bsr, bst, bsb, bsf, lparen, rparen, bslash));
	H_ARULE(octal,	CHX(REP(odigit,3), REP(odigit,2), REP(odigit,1)));
	H_RULE(wrap,	IGN(eol));
	H_RULE(sesc,	h_right(bslash, CHX(escape, octal, wrap, epsilon)));
		/* NB: lone backslashes and escaped newlines are ignored */
	H_ARULE(schars,	h_many(CHX(schar, snest, sesc, eol)));
	H_RULE(snest_,	SEQ(lparen, schars, rparen));
	H_RULE(litstr,	h_middle(TOKD(lparen), schars, rparen));
	H_RULE(hexchr,	h_right(ws, hdigit));
	H_RULE(hexstr,	h_middle(TOKD(langle), h_many(hexchr), TOKD(rangle)));
	H_ARULE(string,	CHX(litstr, hexstr));
	h_bind_indirect(snest, snest_);

	/* arrays and dictionaries */
	H_RULE(dopen,	LIT("<<"));
	H_RULE(dclose,	LIT(">>"));
	H_RULE(obj,	h_indirect());
	H_RULE(k_v,	SEQ(name, obj));
	H_RULE(dict,	h_middle(TOKD(dopen), h_many(k_v), TOKD(dclose)));
	H_RULE(array,	h_middle(TOKD(lbrack), h_many(obj), TOKD(rbrack)));
		// XXX validate: dict keys must be unique

	/* streams */
	H_RULE(stmbeg,	SEQ(dict, KW("stream"), OPT(cr), lf));
	H_RULE(stmend,	SEQ(OPT(eol), LIT("endstream")));
	H_RULE(stream,	h_left(h_bind(stmbeg, kstream, aux), stmend));
		// XXX is whitespace allowed between the eol and "endstream"?

	H_RULE(obj_,	CHX(ref, null, boole, real, intg, name, string,
			    array, dict));
	h_bind_indirect(obj, obj_);

	/*
	 * file structure
	 */

	/* header */
	H_RULE(version,	SEQ(pdigit, IGN(period), pdigit));
	H_RULE(header,	h_middle(LIT("%PDF-"), version, nl));

	/* body */
	H_RULE(indobj,	CHX(stream, obj));
	H_RULE(objdef,	SEQ(pnat, nat, KW("obj"), indobj, KW("endobj")));
	H_RULE(body,	h_many(objdef));	// XXX object streams

	/* cross-reference section */
	H_RULE(xreol,	CHX(SEQ(sp, cr), SEQ(sp, lf), crlf));
		// ^ XXX does the real world follow this rule?! cf. loop.pdf
	H_RULE(xrtyp,	CHX(h_ch('n'), h_ch('f')));
	H_ARULE(xroff,	REP(digit, 10));
	H_ARULE(xrgen,	REP(digit, 5));
	H_ARULE(xrent,	SEQ(xroff, IGN(sp), xrgen, IGN(sp), xrtyp, IGN(xreol)));
	H_ARULE(xrnat,	h_many1(digit));
	H_RULE(xrhead,	SEQ(xrnat, IGN(sp), xrnat, nl));
	H_RULE(xrsub,	SEQ(xrhead, h_many(xrent)));
	H_ARULE(xrefs,	SEQ(KW("xref"), nl, h_many(xrsub)));

	/* cross-reference streams */
	H_RULE(rest,	h_action(h_tell(), act_rest, aux));
	H_RULE(xstream,	h_bind(SEQ(stmbeg, rest), kxstream, aux));
	H_AVRULE(xrstm,	SEQ(pnat, nat, KW("obj"), xstream));
		// XXX skip however much we consumed and check for "endstream endobj"?

	/* trailer */
	H_RULE(startxr,	SEQ(nl, KW("startxref"), nl,
			    lws, xrnat, nl,
			    LIT("%%EOF"), CHX(nl, end)));
		// XXX should lws be allowed before EOF marker?
		// NB: lws before xref offset is allowed, cf. p.48 (example 4)
	H_RULE(xr_td,	SEQ(xrefs, KW("trailer"), dict));

	H_RULE(tail,	SEQ(body, h_optional(xr_td), startxr));
	H_RULE(pdf,	SEQ(header, h_many1(tail), end));

	/* debug parser to consume as much as possible */
	H_RULE(pdfdbg,	SEQ(header, h_many(tail), body, OPT(xr_td), OPT(startxr)));


	/* global parser variables */
	p_pdf = pdf;
	p_pdfdbg = pdfdbg;
	p_startxref = startxr;
	p_xref = CHX(xr_td, xrstm);
	p_objdef = objdef;

	p_fail = h_nothing_p();
	p_epsilon = epsilon;
	p_return_0 = h_action(epsilon, act_return_uint, (void *)0);
	p_return_1 = h_action(epsilon, act_return_uint, (void *)1);
}


/*
 * stream object handling incl. cross-reference streams
 */

#include <limits.h>	/* INT_MAX */
#include <zlib.h>
#include <err.h>

/* combine current position with env=(input,sz) into HBytes */
HParsedToken *
act_ks_bytes(const HParseResult *p, void *env)
{
	const HBytes *bs = env;
	size_t offset = H_CAST_UINT(p->ast) / 8;

	/*
	 * NB: we must allocate a new HBytes struct here because the old one is
	 * allocated only temporarily for the lifetime of the continuation
	 * below.
	 */
	return H_MAKE_BYTES(bs->token + offset, bs->len);
}

XREntry *
lookup_xref(struct Env *aux, size_t nr, size_t gen)
{
	/* for each cross-reference section (i.e. update) */
	for (size_t i = 0; i < aux->nxrefs; i++) {
		HCountedArray *subs = H_INDEX_SEQ(aux->xrefs[i], 0);

		/* for each cross-reference subsection */
		for (size_t j = 0; j < subs->used; j++) {
			HParsedToken *ss = subs->elements[i];
			size_t base = H_INDEX_UINT(ss, 0, 0);
			size_t n = H_INDEX_UINT(ss, 0, 1);

			if (nr >= base && nr - base < n)
				return H_INDEX(XREntry, ss, 1, nr - base);
		}
	}

	return NULL;
}

const HParsedToken *
parse_obj(struct Env *aux, Ref r, size_t offset)
{
	HParseResult *res;
	size_t nr, gen;

	if (offset >= aux->sz) {
		fprintf(stderr, "%s: obj %zu %zu: position %zu (0x%zx) out of "
		    "bounds\n", aux->infile, r.nr, r.gen, offset, offset);
		return NULL;
	}

	res = h_parse(p_objdef, aux->input + offset, aux->sz - offset);
	if (res == NULL) {
		fprintf(stderr, "%s: error parsing object %zu %zu at position "
		    "%zu (0x%zx)\n", aux->infile, r.nr, r.gen, offset, offset);
		return NULL;
	}

	nr = H_INDEX_UINT(res->ast, 0);
	gen = H_INDEX_UINT(res->ast, 1);
	if (nr != r.nr || gen != r.gen) {
		fprintf(stderr, "%s: object ID mismatch at position %zu "
		    "(%#zx): sought %zu %zu, found %zu %zu.\n", aux->infile,
		    offset, offset, r.nr, r.gen, nr, gen);
		return NULL;
	}

	return H_INDEX_TOKEN(res->ast, 2);
}

const HParsedToken *
resolve(struct Env *aux, const HParsedToken *v)
{
	XREntry *entry = NULL;
	Ref *r;

	/* direct objects pass through */
	if (v == NULL || v->token_type != TT_Ref)
		return v;

	/* we are looking at an indirect reference */
	r = v->user;

	/* find the xref entry for this reference */
	entry = lookup_xref(aux, r->nr, r->gen);
	if (entry == NULL)
		return NULL;			/* obj not found */
	if (entry->obj != NULL)
		return resolve(aux, entry->obj);

	/* parse the object and memoize */
	entry->obj = v;				/* break loops */
	switch (entry->type)
	{
	case XR_FREE:
		return NULL;			/* obj deleted */
	case XR_INUSE:
		if (entry->n.gen != r->gen)
			return NULL;		/* obj nr reused */
		entry->obj = parse_obj(aux, *r, entry->n.offs);
		break;
	case XR_OBJSTM:
		if (r->gen != 0)
			return NULL;		/* invalid entry! */
		//XXX entry->obj = parse_obj_stm(aux, entry->o.stm, entry->o.idx);
		//XXX break;
		return NULL;
	}

	return resolve(aux, entry->obj);
}

/*
 * This continuation takes the stream dictionary (as first element of x) and
 * should return a parser that consumes exactly the bytes that make up the
 * stream data.
 */
HParser *
kstream(HAllocator *mm__, const HParsedToken *x, void *env)
{
	struct Env *aux = env;
	const HParsedToken *dict_t = H_INDEX_TOKEN(x, 0);
	const HCountedArray *dict = H_CAST_SEQ(dict_t);
	const HParsedToken *v = NULL;
	size_t sz;

	/* look for the Length entry */
	v = dictentry(dict, "Length");
	v = resolve(aux, v);		/* resolve indirect references */
	if (v == NULL || v->token_type != TT_SINT || v->sint < 0)
		goto fail;
	sz = (size_t)v->sint;

	//fprintf(stderr, "parsing stream object, length %zu.\n", sz);	// XXX debug

	/* dummy struct to hold the pair (input,sz) */
	HBytes *bytes = h_alloc(mm__, sizeof(HBytes));
	bytes->token = aux->input;
	bytes->len = sz;

	HParser *tell = h_tell__m(mm__);
	HParser *skip = h_skip__m(mm__, sz * 8);

	HParser *bytes_p = h_action__m(mm__, tell, act_ks_bytes, bytes);
	HParser *dict_p  = p_return__m(mm__, dict_t);
	return h_sequence__m(mm__, dict_p, bytes_p, skip, NULL);
fail:
#if 0
	if (v == NULL)
		fprintf(stderr, "stream /Length missing\n");
	else if (v -> token_type != TT_SINT)
		fprintf(stderr, "stream /Length not an integer\n");
	else if (v < 0)
		fprintf(stderr, "stream /Length negative\n");
#endif
	//h_pprintln(stderr, p);	// XXX debug
	return p_fail;
}

/*
 * validate the /Type field on a cross-reference stream.
 *
 * p = pnat nat (dict offs offs)
 */
bool
validate_xrstm(HParseResult *p, void *u)
{
	const HCountedArray *tdict = H_FIELD_SEQ(2, 0);
	const HParsedToken *v = dictentry(tdict, "Type");

#if 0
	if (v == NULL)
		fprintf(stderr, "stream dict has no /Type\n");
	else if (v->token_type != TT_BYTES)
		fprintf(stderr, "stream /Type is no name object\n");
	else if (bytes_eq(v->bytes, "XRef"))
		return true;
	return false;
#endif

	return (v != NULL && v->token_type == TT_BYTES &&
	    bytes_eq(v->bytes, "XRef"));
}

struct predictor {
	/* parameters */
	int num;	/* default: 1 (no prediction) */
	int colors;	/* default: 1 */
	int bpc;	/* bits per component; default: 8 */
	int columns;	/* default: 1 */

	int rowsz;	/* bytes per row = ceil(colors * bpc * columns / 8) */

	/* state */
	HSuspendedParser *sp;
	uint8_t	(*predfun)(int, int, int);
	uint8_t	*buf;		/* previous row of input */
	uint8_t	 c;		/* byte 'c' (upper left) */
	int	 x;		/* current position */
};

int
depred_none(struct predictor *pred, uint8_t *inp, size_t sz)
{
	return h_parse_chunk(pred->sp, inp, sz);
}

uint8_t pp_none(int a, int b, int c)	{ return 0; }
uint8_t pp_sub(int a, int b, int c)	{ return a; }
uint8_t pp_up(int a, int b, int c)	{ return b; }
uint8_t pp_avg(int a, int b, int c)	{ return (a + b) / 2; }

#include <stdlib.h>	/* abs() */

uint8_t
pp_paeth(int a, int b, int c)
{
	int p	= a + b - c;
	int pa	= abs(p - a);
	int pb	= abs(p - b);
	int pc	= abs(p - c);

	if (pa <= pb && pa <= pc) return a;
	if (pb <= pc) return b;
	return c;
}

int
depred_png(struct predictor *pred, uint8_t *inp, size_t sz)
{
	/* NB:
	 * at this point, the specific value of pred->num no longer matters.
	 * the PNG predictor tags each row with the function used for that row
	 * and decoding always follows the tag.
	 */
	static uint8_t (*predfuns[])(int, int, int) =
	    {pp_none, pp_sub, pp_up, pp_avg, pp_paeth};

	bool done = false;
	int bpp;

	bpp = (pred->colors * pred->bpc + 7) / 8;	/* bytes per pixel */
	assert (bpp > 0);

	for (size_t i=0; i < sz && !done; i++) {
		int x = pred->x;
		int a = x<bpp ? 0 : pred->buf[x-bpp];	/* left */
		int b = pred->buf[x];			/* up */
		int c = pred->c;			/* up left */

		if (pred->predfun == NULL) {	/* we are before a new row */
			/* select predictor function */
			if (inp[i] > 4) {
				fprintf(stderr, "unknown PNG predictor %d\n",
				    (int)inp[i]);
				return -1;
			}
			pred->predfun = predfuns[inp[i]];

			/* consume the tag */
			if (++i == sz)
				break;
		}

		/* undo the prediction and save the decoded value */
		pred->buf[x] = inp[i] + pred->predfun(a, b, c);

		/* advance to the right */
		pred->c = b;
		pred->x = ++x;

		/* when row complete, pass it to parser and start a new row */
		if (x == pred->rowsz) {
			done = h_parse_chunk(pred->sp, pred->buf, pred->rowsz);
			pred->c = pred->x = 0;
			if (pred->num != 2)	/* support for 8-bpc TIFF */
				pred->predfun = NULL;
		}
	}

	return done;
}

HParseResult *
FlateDecode(HAllocator *mm__, HCountedArray *parms, HBytes b, HParser *p)
{
	size_t const BUFSIZE = 8 * 1024;
	uint8_t *buf;
	HSuspendedParser *sp;
	HParseResult *res;
	const HParsedToken *v;
	size_t sz;
	int done;
	z_stream strm = {0};
	int ret;
	struct predictor pred = {1, 1, 8, 1};
	int (*depredict)(struct predictor *, uint8_t *, size_t);

	/* set up the predictor (if any) */
	#define SETPARM(VAR,STR) do {					\
		v = dictentry(parms, (STR));				\
		if (v != NULL) {					\
			if (v->token_type != TT_SINT || v->sint < 0)	\
				return NULL;				\
			VAR = v->sint;					\
		} } while(0)
	SETPARM(pred.num,	"Predictor");
	SETPARM(pred.colors,	"Colors");
	SETPARM(pred.bpc,	"BitsPerComponent");
	SETPARM(pred.columns,	"Columns");
	#undef SETPARM
	if (pred.num == 1)
		depredict = depred_none;
	else {
		if (pred.num >= 10 && pred.num <= 15)
			depredict = depred_png;
		else if (pred.num == 2) {
			/* for 8-bpc TIFF pred. 2, we can reuse PNG Sub */
			if (pred.bpc == 8) {
				pred.predfun = pp_sub;	/* predict left */
				depredict = depred_png;
			} else {
				// XXX add general TIFF predictor (bpc != 8)
				fprintf(stderr, "FlateDecode: /Predictor %d "
				    "not supported for /BitsPerComponent %d\n",
				    pred.num, pred.bpc);
				return NULL;
			}
		} else {
			fprintf(stderr, "FlateDecode: /Predictor %d"
			    " not supported\n", pred.num);
			return NULL;
		}

		/* allocate row buffer */
		if (pred.columns > (INT_MAX - 7) / pred.colors / pred.bpc) {
			fprintf(stderr, "FlateDecode: overflow\n");
			return NULL;
		}
		pred.rowsz = (pred.colors * pred.bpc * pred.columns + 7) / 8;
		pred.buf = h_alloc(mm__, pred.rowsz);
		memset(pred.buf, 0, pred.rowsz);
	}

	/* set up zlib */
	// XXX pass our allocator to zlib
	ret = inflateInit(&strm);
	if (ret != Z_OK)
		errx(1, "inflateInit: %s (%d)", strm.msg, ret);
	buf = h_alloc(mm__, BUFSIZE);

	/* initialize target parser */
	sp = h_parse_start__m(mm__, p);
	assert(sp != NULL);
	pred.sp = sp;

	done = 0;
	strm.avail_in = b.len;
	strm.next_in = (unsigned char *)b.token;
	do {
		strm.avail_out = BUFSIZE;
		strm.next_out = buf;

		ret = inflate(&strm, Z_NO_FLUSH);
		if (ret != Z_STREAM_END && ret != Z_OK) {
			fprintf(stderr, "inflate: %s (%d)\n", strm.msg, ret);
			break;
		}

		sz = BUFSIZE - strm.avail_out;
		done = depredict(&pred, buf, sz);
	} while (done == 0 && ret == Z_OK);

	res = h_parse_finish(sp);
		// XXX always return NULL on error?
	inflateEnd(&strm);
	mm__->free(mm__, pred.buf);
	mm__->free(mm__, buf);

	if (done == -1)
		return NULL;
	return res;
}

/*
 * decode the byte stream 'b' according to metadata in its stream dictionary
 * 'd' and parse the result with 'p'.
 */
HParseResult *
parse_stream(HAllocator  *mm__, HCountedArray *d, HBytes b, HParser *p)
{
	HParseResult *(*filter)(HAllocator *, HCountedArray *, HBytes, HParser *);
	HCountedArray *parms = NULL;
	const HParsedToken *v;

	v = dictentry(d, "Filter");
	if (v == NULL)
		return h_parse__m(mm__, p, b.token, b.len);

	/* compile to a CF backend to enable incremental parsing */
	if (h_compile(p, PB_LLk, NULL) == -1)
		errx(1, "xref data parser: LL(1) compile failed");

	if (v->token_type == TT_SEQUENCE)
		return NULL;	// XXX filter chains not supported, yet
	assert(v->token_type == TT_BYTES);
	if (bytes_eq(v->bytes, "FlateDecode"))
		filter = FlateDecode;
	else
		return NULL;		/* filter not supported */

	v = dictentry(d, "DecodeParms");
	if (v && v->token_type == TT_SEQUENCE)
		parms = v->seq;

	return filter(mm__, parms, b, p);
}

/*
 * interpret a cross-reference stream and return it in the same form as other
 * cross-reference sections:
 *
 * p = (pnat nat (dict xrefs))
 * result = (xrefs dict)
 */
HParsedToken *
act_xrstm(const HParseResult *p, void *u)
{
	HParsedToken *xrefs, *dict, *result;

	dict = H_INDEX_TOKEN(p->ast, 2, 0);
	xrefs = H_INDEX_TOKEN(p->ast, 2, 1);

	result = H_MAKE_SEQN(2);
	result->seq->elements[0] = xrefs;
	result->seq->elements[1] = dict;
	result->seq->used = 2;
	return result;
}

HParser *
p_xrefsub__m(HAllocator *mm__, size_t base, size_t count, HParser *p_entry)
{
	HParser *ret_base, *ret_count, *p_header, *p_entries;

	ret_base  = p_return_uint__m(mm__, base);
	ret_count = p_return_uint__m(mm__, count);
	p_header  = h_sequence__m(mm__, ret_base, ret_count, NULL);
	p_entries = h_repeat_n__m(mm__, p_entry, count);

	return h_sequence__m(mm__, p_header, p_entries, NULL);
}

/* x = ((dict ...) bytes) */
HParser *
kxstream(HAllocator *mm__, const HParsedToken *x, void *env)
{
	//struct Env *aux = env;
	const HParsedToken *v, *dict_t;
	const HParseResult *res;
	HCountedArray *dict;
	HBytes bytes;
	size_t W[3];
	size_t Size, Wn, Wskip;
	HParser *p_field[3], *p_entry, **p_subs, *p_xrefdata;

	dict_t = H_INDEX_TOKEN(x, 0, 0);
	dict = H_CAST_SEQ(dict_t);
	bytes = H_INDEX_BYTES(x, 1);

	/*
	 * what follows is a horrible bunch of code that builds, from the
	 * entries /W, /Index, and /Size in the stream dictionary, a parser for
	 * the cross-reference data itself.
	 *
	 * in short, every cross-reference entry consists of (as of PDF 1.7)
	 * three fields, but it could be more. /W gives the widths (in bytes)
	 * of these fields. the /Index specifies the division of the data into
	 * subsections; it is an array of natural numbers that in pairs specify
	 * the base object number and length of each subsection - analogous to
	 * the subsection headers in classic xref sections.
	 *
	 * when /Index is missing, a default value of [0 Size] is defined,
	 * where Size is the value of the /Size field. as in normal trailer
	 * dictionaries, it specifies the total size of the (entire)
	 * cross-reference table.
	 *
	 * when /W states a width of 0 for a field, that field is not present
	 * in the data and a default value should be used "if there is one".
	 * most notably, the first field determines the "type" of the entry,
	 * analogous to the 'n' and 'f' tags in classic xref sections; a width
	 * of 0 for the first field is specified to mean that every entry is of
	 * type 1 (= "n"). that type, in particular, specifies a default of 0
	 * for field 3 (generation). in fact, these are the only defaults
	 * defined by ISO 32000-1:2008 (PDF 1.7).
	 *
	 *   entry type  field no.  default value
	 *               1 (type)   1
	 *   1 ("n")     3 (gen.)   0
	 */

	/* Size (required) - total size of xref table */
	v = dictentry(dict, "Size");
	if (v == NULL || v->token_type != TT_SINT || v->sint < 1)
		goto fail;
	Size = v->sint;

	/* W (required) - field widths for each xref entry */
	v = dictentry(dict, "W");
	if (v == NULL || v->token_type != TT_SEQUENCE)
		goto fail;
	if ((Wn = v->seq->used) < 3)
		goto fail;
	Wskip = 0;
	for (size_t i = 0; i < Wn; i++) {
		HTokenType tt = v->seq->elements[i]->token_type;
		int64_t w = v->seq->elements[i]->sint;

		if (tt != TT_SINT || w < 0)
			goto fail;
		if (i < 3) {
			/* we can't take >64 bits and want to use size_t */
			if (w > 8 || w > sizeof(size_t))
				goto fail;
			W[i] = (size_t)w;
		} else {
			if (w > SIZE_MAX - Wskip)
				goto fail;	/* overflow */
			Wskip += w;
		}
	}
	if (Wskip > SIZE_MAX / 8)
		goto fail;

	/*
	 * build the parser for one xref entry.
	 *
	 * in summary, the only sensible forms for /W are:
	 *
	 *   [t x y] with t,x,y > 0  full general form
	 *   [0 x y] with x,y > 0    only type-1 ("in use") entries
	 *   [0 x 0] with x > 0      only type-1 entries, only offsets
	 *
	 * however, though nonsensical, [t x 0] with t,x > 0 is not disallowed
	 * by the spec; as long as all entries are of type 1, the xref data can
	 * be interpreted without ambiguity.
	 *
	 * in fact, every nonsensical form is possible as long as there are 0
	 * entries.
	 *
	 * we realize this mess by just initializing the default parser to
	 * p_fail and and replacing the known cases afterwards.
	 */
	for (size_t i = 0; i < 3; i++) {
		if (W[i] == 0)
			p_field[i] = p_fail;	/* no known default */
		else
			p_field[i] = h_bits__m(mm__, W[i] * 8, false);
	}
	/* known default cases: */
	if (W[0] == 0)
		p_field[0] = p_return_1;	/* all type 1 */
	if (W[2] == 0) {
		p_field[2] = p_return_0;	/* all generation 0 */
		if (W[0] > 0) {
			/* type field *must* be 1 */
			p_field[0] = h_attr_bool__m(mm__, p_field[0],
			    validate_eq_uint, (void *)1);
		}
	}
	if (Wskip > 0)	// XXX h_skip does not work with CF, yet
		goto fail;
	p_entry = h_sequence__m(mm__, p_field[0], p_field[1], p_field[2], NULL);
	p_entry = h_action__m(mm__, p_entry, act_xrstment, NULL);

	/* Index (optional) - subsections [base count ...] */
	v = dictentry(dict, "Index");
	if (v == NULL) {
		/* default: [0 Size] */
		p_subs = h_alloc(mm__, 2 * sizeof(HParser *));
		p_subs[0] = p_xrefsub__m(mm__, 0, Size, p_entry);
		p_subs[1] = NULL;
	} else if (v->token_type != TT_SEQUENCE) {
		goto fail;
	} else {
		size_t nsubs = v->seq->used / 2;

		/* build a parser for each subsection */
		if (nsubs >= SIZE_MAX / sizeof(HParser *))
			goto fail;
		p_subs = h_alloc(mm__, (nsubs + 1) * sizeof(HParser *));
		for (size_t i = 0; i < nsubs; i++) {
			HParsedToken *base = v->seq->elements[2 * i];
			HParsedToken *n = v->seq->elements[2 * i + 1];

			if (base->token_type != TT_SINT || base->sint < 0 ||
			    n->token_type != TT_SINT || n->sint < 0 ||
			    n->sint > SIZE_MAX)
				goto fail;

			p_subs[i] = p_xrefsub__m(mm__, base->sint, n->sint,
			    p_entry);
		}
		p_subs[nsubs] = NULL;
	}
	p_xrefdata = h_sequence__ma(mm__, (void **)p_subs);

	/* restrict bytes to Length if present (and not indirect) */
	v = dictentry(dict, "Length");
	if (v != NULL && v->token_type == TT_SINT && v->sint >= 0)
		bytes.len = v->sint;

	/* decode and parse the stream data */
	res = parse_stream(mm__, dict, bytes, p_xrefdata);
	if (res == NULL)
		goto fail;

	HParser *dict_p = p_return__m(mm__, dict_t);
	HParser *xref_p = p_return__m(mm__, res->ast);
	HParser *skip_p = h_skip__m(mm__, bytes.len * 8);
		// XXX skip only as much as parse_stream consumed

	return h_sequence__m(mm__, dict_p, xref_p, skip_p, NULL);
fail:
	return p_fail;
}


/*
 * main program
 */

#include <stdio.h>
#include <inttypes.h>
#include <stdlib.h>	/* realloc() */
#include <fcntl.h>	/* open() */
#include <unistd.h>	/* lseek() */
#include <sys/mman.h>	/* mmap() */

const char *infile = NULL;

/*
 * This helper implements the standard backwards parsing strategy to read all
 * cross-reference sections and trailer dictionaries, starting from the
 * 'startxref' offset found at the very end of the input.
 *
 * Allocates and returns an array of HParsedTokens, each containing the result
 * of a successful 'p_xref' parse. Sets the output parameter 'nxrefs' to the
 * number of elements.
 *
 * A return value of NULL indicates an empty result.
 */
const HParsedToken **
parse_xrefs(const char *input, size_t sz, size_t *nxrefs)
{
	HParseResult *res = NULL;
	const HParsedToken **xrefs = NULL;	/* empty result */
	const HParsedToken *tok = NULL;
	size_t n = 0;
	size_t offset = 0;

	// XXX try formulating this as a parser using h_seek()

	/* search for the "startxref" section from the back of the file */
	HParser *p = h_left(p_startxref, h_end_p());
	for (size_t i = 0; i < sz; i++) {
		res = h_parse(p, input + sz - i, i);
		if (res != NULL)
			break;
	}
	if (res == NULL) {
		fprintf(stderr, "%s: startxref not found\n", infile);
		goto end;
	}
	offset = H_INDEX_UINT(res->ast, 0);

	for (;;) {
		//res = h_parse(p_xref, input + offset, sz - offset);
		HParser *p = h_right(h_seek(offset * 8, SEEK_SET), p_xref);	// XXX
		res = h_parse(p, input, sz);
		if (res == NULL) {
			fprintf(stderr, "%s: error parsing xref section at "
			    "position %zu (0x%zx)\n", infile, offset, offset);
			break;
		}

		/* save this section in xrefs */
		if (n >= SIZE_MAX / sizeof(HParsedToken *))
			errx(1, "realloc: size would overflow");
		xrefs = realloc(xrefs, (n + 1) * sizeof(HParsedToken *));
		if (xrefs == NULL)
			err(1, "realloc");
		xrefs[n++] = res->ast;

		/* look up the next offset (to the previous xref section) */
		tok = dictentry(H_INDEX_SEQ(res->ast, 1), "Prev");
		if (tok == NULL)
			break;
		if (tok->token_type != TT_SINT) {
			fprintf(stderr, "%s: /Prev not an integer\n", infile);
			break;
		}

		/*
		 * validate the new offset. we don't want to get caught in a
		 * loop. the offsets should strictly decrease, unless the file
		 * is a "linearized" PDF. in that case there should be exactly
		 * two xref sections in the reverse order, so we allow the
		 * first section to point forward.
		 */
		if (n > 1 && tok->sint >= offset) {
			fprintf(stderr, "%s: /Prev pointer of xref section at "
			    "%zu (0x%zx) points forward\n", infile, offset,
			    offset);
			break;
		}

		offset = (size_t)tok->sint;
	}

end:
	*nxrefs = n;
	return xrefs;
}

int
main(int argc, char *argv[])
{
	struct Env aux;
	HParseResult *res = NULL;
	const uint8_t *input;
	size_t sz;
	int fd;

	/* command line handling */
	if (argc != 2) {
		fprintf(stderr, "usage: %s file\n", argv[0]);
		return 1;
	}
	infile = argv[1];

	/* mmap the input file */
	fd = open(infile, O_RDONLY);
	if (fd == -1)
		err(1, "%s", infile);
	sz = lseek(fd, 0, SEEK_END);
	if (sz == -1)
		err(1, "lseek");
	input = mmap(NULL, sz?sz:1, PROT_READ, MAP_PRIVATE, fd, 0);
	if (input == MAP_FAILED)
		err(1, "mmap");

	/* build parsers */
	aux = (struct Env){infile, input, sz};
	init_parser(&aux);

	/* parse all cross-reference sections and trailer dictionaries */
	aux.xrefs = parse_xrefs(input, sz, &aux.nxrefs);
	// XXX debug
	//fprintf(stderr, "%s: %zu xref sections parsed\n", infile, aux.nxrefs);
	//for (size_t i = 0; i < aux.nxrefs; i++)
	//	h_pprintln(stderr, aux.xrefs[i]);

	/* run the main parser */
	res = h_parse(p_pdf, input, sz);
	if (!res) {
		fprintf(stderr, "%s: no parse\n", infile);

		/* help us find the error */
		res = h_parse(p_pdfdbg, input, sz);
		if (res) {
			int64_t pos = res->bit_length / 8;
			fprintf(stderr, "%s: error after position"
			    " %" PRId64 " (0x%" PRIx64 ")\n",
			    infile, pos, pos);
			//h_pprintln(stderr, res->ast);	// XXX debug
		}

		return 1;
	}

	/* print result */
	h_pprintln(stdout, res->ast);

	return 0;
}
